import importlib
from datetime import datetime
import logging
import logging.config
from pathlib import Path
import os

log_levels = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}


class TensorboardWriter:
    def __init__(self, log_dir, logger, enabled):
        self.writer = None
        self.selected_module = ""

        if enabled:
            log_dir = str(log_dir)

            # Retrieve vizualization writer.
            succeeded = False
            for module in ["torch.utils.tensorboard", "tensorboardX"]:
                try:
                    self.writer = importlib.import_module(module).SummaryWriter(log_dir)
                    succeeded = True
                    break
                except ImportError:
                    succeeded = False
                self.selected_module = module

            if not succeeded:
                message = (
                    "Warning: visualization (Tensorboard) is configured to use, but currently not installed on "
                    "this machine. Please install TensorboardX with 'pip install tensorboardx', upgrade PyTorch to "
                    "version >= 1.1 to use 'torch.utils.tensorboard' or turn off the option in the 'config.yaml' file."
                )
                logger.warning(message)

        self.step = 0
        self.mode = ""

        self.tb_writer_ftns = {
            "add_scalar",
            "add_scalars",
            "add_image",
            "add_images",
            "add_audio",
            "add_text",
            "add_histogram",
            "add_pr_curve",
            "add_embedding",
        }
        self.tag_mode_exceptions = {"add_histogram", "add_embedding"}
        self.timer = datetime.now()

    def set_step(self, step, mode="train"):
        self.mode = mode
        self.step = step
        if step == 0:
            self.timer = datetime.now()
        else:
            duration = datetime.now() - self.timer
            self.add_scalar("steps_per_sec", 1 / duration.total_seconds())
            self.timer = datetime.now()

    def __getattr__(self, name):
        """
        If visualization is configured to use:
            return add_data() methods of tensorboard with additional information (step, tag) added.
        Otherwise:
            return a blank function handle that does nothing
        """
        if name in self.tb_writer_ftns:
            add_data = getattr(self.writer, name, None)

            def wrapper(tag, data, *args, **kwargs):
                if add_data is not None:
                    # add mode(train/valid) tag
                    if name not in self.tag_mode_exceptions:
                        tag = "{}/{}".format(tag, self.mode)
                    add_data(tag, data, self.step, *args, **kwargs)

            return wrapper
        else:
            # default action for returning methods defined in this class, set_step() for instance.
            try:
                attr = object.__getattr__(name)
            except AttributeError:
                raise AttributeError(
                    "type object '{}' has no attribute '{}'".format(
                        self.selected_module, name
                    )
                )
            return attr


# FIXME: dictConfig is not loading
def setup_logging(log_config=None, default_level=logging.INFO):
    """
    Setup logging configuration
    """
    save_dir = log_config["log_dir"]
    os.makedirs(save_dir, exist_ok=True)
    if log_config:
        for _, handler in log_config["handlers"].items():
            if "filename" in handler:
                handler["filename"] = str(Path(save_dir) / handler["filename"])
        logging.config.dictConfig(log_config)
    else:
        print(
            "Warning: logging configuration file is not found in {}.".format(log_config)
        )
        logging.basicConfig(level=default_level)


def get_logger(name, verbosity=2):
    msg_verbosity = "verbosity option {} is invalid. Valid options are {}.".format(
        verbosity, log_levels.keys()
    )
    assert verbosity in log_levels, msg_verbosity
    logger = logging.getLogger(name)
    logger.setLevel(log_levels[verbosity])
    return logger



import torch


def accuracy(output, target):
    with torch.no_grad():
        pred = torch.argmax(output, dim=1)
        assert pred.shape[0] == len(target)
        correct = 0
        correct += torch.sum(pred == target).item()
    return correct / len(target)


def top_k_acc(output, target, k=3):
    with torch.no_grad():
        pred = torch.topk(output, k, dim=1)[1]
        assert pred.shape[0] == len(target)
        correct = 0
        for i in range(k):
            correct += torch.sum(pred[:, i] == target).item()
    return correct / len(target)


# mean squared error
def mse(output, target):
    with torch.no_grad():
        batch_size = target.size(0)
        diff = torch.square(output - target) / batch_size
        diff = diff.sum()
    return diff


# mean absolute error
def mae(output, target):
    with torch.no_grad():
        batch_size = target.size(0)
        diff = torch.abs(output - target) / batch_size
        diff = diff.sum()
    return diff



import torch
from torch import nn
from torch.utils.data import Dataset
from typing import Iterable, Dict, Optional, List
from torch import LongTensor
from jac_nlp.bi_ner.model.ph_model import (
    prepare_inputs,
    collate_examples,
    ph_init,
    get_embeddings,
    get_scores,
)
from jac_nlp.bi_ner.model.loss import ContrastiveThresholdLoss
from jac_nlp.bi_ner.model.tokenize_data import get_datasets
from jac_nlp.bi_ner.datamodel.utils import get_category_id_mapping, invert
from jac_nlp.bi_ner.datamodel.example import (
    Example,
    batch_examples,
    BatchedExamples,
    TypedSpan,
)
from functools import partial
from torch.nn.functional import pad
from jac_misc.ph.utils.base import BaseInference

from typing import Any
from collections import defaultdict
import json


def collate_fn(
    examples: Iterable[Example],
    _max_sequence_length=128,
    return_batch_examples: bool = False,
) -> Dict[str, Optional[LongTensor]]:
    return collate_examples(
        examples,
        padding_token_id=100,
        pad_length=_max_sequence_length,
        return_batch_examples=return_batch_examples,
    )


class CustomLoss(torch.nn.Module):
    def __init__(self, n_classes=2, beta=0.6):
        super(CustomLoss, self).__init__()
        self._loss_fn = ContrastiveThresholdLoss(
            n_classes=n_classes, ignore_id=-100, reduction="mean", beta=beta
        )
        self._span_coef = 0.6
        self._start_coef = 0.2
        self._end_coef = 0.2
        self._max_entity_length = 30

    def forward(self, output, labels):
        span_loss = self._loss_fn(output[0], labels)
        start_loss = self._loss_fn(
            output[1].unsqueeze(-2).repeat(1, 1, self._max_entity_length, 1), labels
        )
        end_loss = self._loss_fn(
            output[2].unsqueeze(-2).repeat(1, 1, self._max_entity_length, 1), labels
        )

        return (
            self._span_coef * span_loss
            + self._start_coef * start_loss
            + self._end_coef * end_loss
        )


class CustomModel(nn.Module):
    def __init__(self, model_args) -> None:
        super(CustomModel, self).__init__()
        ph_init(model_args)
        # print(f"in custom model{model_args}")
        con_encoder_layer = nn.TransformerEncoderLayer(
            d_model=768,
            nhead=12,
            dim_feedforward=128,
            batch_first=True,
        )
        cand_encoder_layer = nn.TransformerEncoderLayer(
            d_model=768,
            nhead=12,
            dim_feedforward=128,
            batch_first=True,
        )
        self.con_encoder = nn.TransformerEncoder(
            encoder_layer=con_encoder_layer, num_layers=1
        )
        self.cand_encoder = nn.TransformerEncoder(
            encoder_layer=cand_encoder_layer, num_layers=1
        )
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.con_encoder.to(device)
        self.cand_encoder.to(device)

    def forward(self, x):
        mode = "train"
        if x.ndimension() == 1:
            x = x.unsqueeze(0)
            mode = None
        ent_emb, token_emb = get_embeddings(x)
        token_emb = self.con_encoder(token_emb)
        ent_emb = self.cand_encoder(ent_emb)
        scores = get_scores(token_emb, ent_emb, mode)
        return scores


class CustomDataset(Dataset):
    def __init__(self, train_args) -> None:
        super(CustomDataset, self).__init__()
        with open(train_args["data_file"], "r") as fp:
            self.data = json.load(fp)
        category_id_mapping = get_category_id_mapping(
            train_args, train_args["descriptions"]
        )
        self.example_encoder = partial(
            prepare_inputs,
            category_mapping=invert(category_id_mapping),
            no_entity_category=train_args["unk_category"],
        )
        self.dataset = get_datasets(self.data, self.example_encoder)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        # print(f"dataset :  {self.dataset[idx]}")
        # _,_,input_ids,_,_,labels=self.dataset[idx]
        # return input_ids,labels
        return self.dataset[idx]


class CustomInference(BaseInference):
    @torch.no_grad()
    def predict(self, data: Any) -> Any:
        self._max_entity_length = 30
        category_id_mapping = get_category_id_mapping(
            data["inf_args"], data["inf_args"]["descriptions"]
        )
        category_mapping = invert(category_id_mapping)
        self._no_entity_category = data["inf_args"]["unk_category"]
        self._no_entity_category_id = data["inf_args"]["unk_entity_type_id"]
        self.stride = 0.9
        self._stride_length = int(data["inf_args"]["max_sequence_length"] * self.stride)
        # tokenization and data transformation
        self._text_length: List[Optional[int]] = [None] * len(data["dataset"])
        examples = list(
            prepare_inputs(
                data["dataset"],
                [None] * len(data["dataset"]),
                category_mapping=category_mapping,
                no_entity_category=self._no_entity_category,
                stride=self.stride,
                text_lengths=self._text_length,
            )
        )
        example = batch_examples(
            examples,
            batch_size=1,
            collate_fn=partial(collate_fn, return_batch_examples=True),
        )
        predictions_collector = [defaultdict(int) for _ in data["dataset"]]
        for batch in example:
            input_data = batch["input_ids"]
            input_data = input_data.to(self.device)
            predictions: LongTensor = self.model(input_data[0])
            batched_examples: BatchedExamples = batch["examples"]

            batch_size, length = batched_examples.start_offset.shape
            span_start = (
                pad(
                    batched_examples.start_offset,
                    [0, data["inf_args"]["max_sequence_length"] - length],
                    value=-100,
                )
                .view(batch_size, data["inf_args"]["max_sequence_length"], 1)
                .repeat(1, 1, self._max_entity_length)
                .to(self.device)
            )

            end_offset = pad(
                batched_examples.end_offset,
                [0, data["inf_args"]["max_sequence_length"] - length],
                value=-100,
            ).to(self.device)

            padding_masks = []
            span_end = []
            for shift in range(
                self._max_entity_length
            ):  # self._max_entity_length ~ 20-30, so it is fine to not vectorize this
                span_end.append(torch.roll(end_offset, -shift, 1).unsqueeze(-1))
                padding_mask = torch.roll(end_offset != -100, -shift, 1)
                padding_mask[:, -shift:] = False
                padding_masks.append(padding_mask.unsqueeze(-1))

            span_end = torch.concat(span_end, dim=-1)
            padding_mask = torch.concat(padding_masks, dim=-1)
            entities_mask = (
                (predictions != self._no_entity_category_id)
                # & padding_mask
                & (span_end != -100)
                & (span_start != -100)
            )
            entity_token_start = (
                torch.arange(data["inf_args"]["max_sequence_length"])
                .reshape(1, data["inf_args"]["max_sequence_length"], 1)
                .repeat(batch_size, 1, self._max_entity_length)
                .to(self.device)
            )

            entity_text_ids = (
                torch.tensor(batched_examples.text_ids)
                .view(batch_size, 1, 1)
                .repeat(
                    1, data["inf_args"]["max_sequence_length"], self._max_entity_length
                )
                .to(self.device)
            )

            chosen_text_ids = entity_text_ids[entities_mask]
            chosen_category_ids = predictions[entities_mask]
            chosen_span_starts = span_start[entities_mask]
            chosen_span_ends = span_end[entities_mask]
            chosen_token_starts = entity_token_start[entities_mask]
            for text_id, category_id, start, end, token_start in zip(
                chosen_text_ids,
                chosen_category_ids,
                chosen_span_starts,
                chosen_span_ends,
                chosen_token_starts,
            ):
                predictions_collector[text_id][
                    (
                        TypedSpan(
                            start.item(),
                            end.item(),
                            category_id_mapping[category_id.item()],
                        ),
                        token_start.item(),
                    )
                ] += 1
        all_entities = [set() for _ in data["dataset"]]
        for text_id, preds in enumerate(predictions_collector):
            text_length = self._text_length[text_id]
            strided_text_length = (
                (text_length // self._stride_length)
                + (text_length % self._stride_length > 0)
            ) * self._stride_length
            for (entity, entity_token_start), count_preds in preds.items():
                # [1, 2, 3, ..., MAX, MAX, ..., MAX, MAX - 1, ..., 3, 2, 1]
                #  bin sizes are stride_length except for the last bin
                total_predictions = min(
                    (entity_token_start // self._stride_length) + 1,
                    (
                        max(
                            strided_text_length
                            - data["inf_args"]["max_sequence_length"],
                            0,
                        )
                        // self._stride_length
                    )
                    + 1,
                    ((strided_text_length - entity_token_start) // self._stride_length)
                    + 1,
                )
                if count_preds >= total_predictions // 2:
                    all_entities[text_id].add(entity)
        return all_entities



from jaseci.jsorc.live_actions import jaseci_action


@jaseci_action(act_group=["example_module"], allow_remote=True)
def call(a: str):
    return f"example module received {a}."


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



from transformers import MBartForConditionalGeneration, MBart50TokenizerFast
from jaseci.jsorc.live_actions import jaseci_action
from typing import Union, List
from fastapi import HTTPException
import traceback

model = None
tokenizer = None


@jaseci_action(act_group=["translator"], allow_remote=True)
def setup():
    global model, tokenizer, supported_languages
    model = MBartForConditionalGeneration.from_pretrained(
        "facebook/mbart-large-50-many-to-many-mmt"
    )
    tokenizer = MBart50TokenizerFast.from_pretrained(
        "facebook/mbart-large-50-many-to-many-mmt"
    )
    supported_languages = list(tokenizer.lang_code_to_id.keys())


setup()


@jaseci_action(act_group=["translator"], allow_remote=True)
def translate(text: Union[str, List[str]], src_lang: str, tgt_lang: str) -> List[str]:
    """
    Translate text from one language to another.

    Args:
        text (Union[str, List[str]]): Text to translate.
        src_lang (str): Source language.
        tgt_lang (str): Target language.

    Returns:
        List[str]: Translated text.
    """
    global model, tokenizer
    try:
        if src_lang not in supported_languages:
            raise ValueError(f"Unsupported source language: {src_lang}")
        if tgt_lang not in supported_languages:
            raise ValueError(f"Unsupported target language: {tgt_lang}")

        if isinstance(text, str):
            text = [text]
        tokenizer.src_lang = src_lang
        forced_bos_token_id = tokenizer.lang_code_to_id[tgt_lang]
        encoded = tokenizer(text, return_tensors="pt")
        generated_tokens = model.generate(
            **encoded,
            forced_bos_token_id=forced_bos_token_id,
        )
        return tokenizer.batch_decode(generated_tokens, skip_special_tokens=True)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["translator"], allow_remote=True)
def get_supported_languages() -> List[str]:
    """
    Get a list of supported languages.

    Returns:
        List[str]: List of supported languages.
    """
    return supported_languages



import json
import torch
import pandas as pd
from pathlib import Path
from itertools import repeat
from collections import OrderedDict, abc
import yaml


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)


def read_json(fname):
    fname = Path(fname)
    with fname.open("rt") as handle:
        return json.load(handle, object_hook=OrderedDict)


def read_yaml(fname):
    fname = Path(fname)
    with fname.open("rt") as handle:
        return yaml.load(handle, Loader=yaml.FullLoader)


def write_yaml(content, fname):
    fname = Path(fname)
    with fname.open("wt") as handle:
        yaml.dump(content, handle, default_flow_style=False)


def write_json(content, fname):
    fname = Path(fname)
    with fname.open("wt") as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def inf_loop(data_loader):
    """wrapper function for endless data loader."""
    for loader in repeat(data_loader):
        yield from loader


def deep_update(d, u):
    for k, v in u.items():
        if isinstance(v, abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def prepare_device(n_gpu_use):
    """
    setup GPU device if available. get gpu device indices which are used for DataParallel
    """
    n_gpu = torch.cuda.device_count()
    if n_gpu_use > 0 and n_gpu == 0:
        print(
            "Warning: There's no GPU available on this machine,"
            "training will be performed on CPU."
        )
        n_gpu_use = 0
    if n_gpu_use > n_gpu:
        print(
            f"Warning: The number of GPU's configured to use is {n_gpu_use}, but only {n_gpu} are "
            "available on this machine."
        )
        n_gpu_use = n_gpu
    device = torch.device("cuda:0" if n_gpu_use > 0 else "cpu")
    list_ids = list(range(n_gpu_use))
    return device, list_ids


def save_custom_python(python: str, path: str = "heads/custom.py"):
    with open(path, "w") as f:
        f.write(python)


class MetricTracker:
    def __init__(self, *keys, writer=None):
        self.writer = writer
        self._data = pd.DataFrame(index=keys, columns=["total", "counts", "average"])
        self.reset()

    def reset(self):
        for col in self._data.columns:
            self._data[col].values[:] = 0

    def update(self, key, value, n=1):
        if self.writer is not None:
            self.writer.add_scalar(key, value)
        self._data.total[key] += value * n
        self._data.counts[key] += n
        self._data.average[key] = self._data.total[key] / self._data.counts[key]

    def avg(self, key):
        return self._data.average[key]

    def result(self):
        return dict(self._data.average)



import numpy as np
import torch

from .base import BaseTrainer
from .util import inf_loop, MetricTracker


class Trainer(BaseTrainer):
    """
    Trainer class
    """

    def __init__(
        self,
        model,
        criterion,
        metric_ftns,
        optimizer,
        config,
        device,
        data_loader,
        uuid,
        valid_data_loader=None,
        lr_scheduler=None,
        len_epoch=None,
        resume=None,
    ):
        super().__init__(model, criterion, metric_ftns, optimizer, config, uuid, resume)
        self.config = config
        self.device = device
        self.data_loader = data_loader
        if len_epoch is None:
            # epoch-based training
            self.len_epoch = len(self.data_loader)
        else:
            # iteration-based training
            self.data_loader = inf_loop(data_loader)
            self.len_epoch = len_epoch
        self.valid_data_loader = valid_data_loader
        self.do_validation = self.valid_data_loader is not None
        self.lr_scheduler = lr_scheduler
        self.log_step = int(np.sqrt(data_loader.batch_size))

        self.train_metrics = MetricTracker(
            "loss", *[m.__name__ for m in self.metric_ftns], writer=self.writer
        )
        self.valid_metrics = MetricTracker(
            "loss", *[m.__name__ for m in self.metric_ftns], writer=self.writer
        )

    def _train_epoch(self, epoch):
        """
        Training logic for an epoch
        :param epoch: Integer, current training epoch.
        :return: A log that contains average loss and metric in this epoch.
        """
        self.model.train()
        self.train_metrics.reset()

        for batch_idx, (data, target) in enumerate(self.data_loader):
            data, target = data.to(self.device), target.to(self.device)

            self.optimizer.zero_grad()
            output = self.model(data)
            loss = self.criterion(output, target)
            loss.backward()
            self.optimizer.step()

            self.writer.set_step((epoch - 1) * self.len_epoch + batch_idx)
            self.train_metrics.update("loss", loss.item())
            for met in self.metric_ftns:
                self.train_metrics.update(met.__name__, met(output, target))

            if batch_idx % self.log_step == 0:
                self.logger.debug(
                    "Train Epoch: {} {} Loss: {:.6f}".format(
                        epoch, self._progress(batch_idx), loss.item()
                    )
                )
                # self.writer.add_image('input', make_grid(
                #     data.cpu(), nrow=8, normalize=True))

            if batch_idx == self.len_epoch:
                break
        log = self.train_metrics.result()

        if self.do_validation:
            val_log = self._valid_epoch(epoch)
            log.update(**{"val_" + k: v for k, v in val_log.items()})

        if self.lr_scheduler is not None:
            self.lr_scheduler.step()
        return log

    def _valid_epoch(self, epoch):
        """
        Validate after training an epoch
        :param epoch: Integer, current training epoch.
        :return: A log that contains information about validation
        """
        self.model.eval()
        self.valid_metrics.reset()
        with torch.no_grad():
            for batch_idx, (data, target) in enumerate(self.valid_data_loader):
                data, target = data.to(self.device), target.to(self.device)

                output = self.model(data)
                loss = self.criterion(output, target)

                self.writer.set_step(
                    (epoch - 1) * len(self.valid_data_loader) + batch_idx, "valid"
                )
                self.valid_metrics.update("loss", loss.item())
                for met in self.metric_ftns:
                    self.valid_metrics.update(met.__name__, met(output, target))
                # self.writer.add_image('input', make_grid(
                #     data.cpu(), nrow=8, normalize=True))

        # add histogram of model parameters to the tensorboard
        for name, p in self.model.named_parameters():
            self.writer.add_histogram(name, p, bins="auto")
        return self.valid_metrics.result()

    def _progress(self, batch_idx):
        base = "[{}/{} ({:.0f}%)]"
        if hasattr(self.data_loader, "n_samples"):
            current = batch_idx * self.data_loader.batch_size
            total = self.data_loader.n_samples
        else:
            current = batch_idx
            total = self.len_epoch
        return base.format(current, total, 100.0 * current / total)



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.jsorc.live_actions import load_module_actions, unload_module


class PDFExtModule(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(PDFExtModule, cls).setUpClass()
        ret = load_module_actions("jac_misc.pdf_ext")
        assert ret == True

    @jac_testcase("pdf_ext.jac", "test_valid_pdf_url")
    def test_valid_pdf_url(self, ret):
        self.assertIn("content", ret["report"][0].keys())

    @jac_testcase("pdf_ext.jac", "test_metadata_enabled")
    def test_metadata_enabled(self, ret):
        self.assertIn("metadata", ret["report"][0].keys())

    @jac_testcase("pdf_ext.jac", "test_metadata_disabled")
    def test_metadata_disabled(self, ret):
        self.assertNotIn("metadata", ret["report"][0].keys())

    @classmethod
    def tearDownClass(cls):
        super(PDFExtModule, cls).tearDownClass()
        ret = unload_module("jac_misc.pdf_ext.pdf_ext")
        assert ret == True



import pytest

from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.jsorc.live_actions import load_module_actions, unload_module


class TextClusterModule(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(TextClusterModule, cls).setUpClass()
        ret = load_module_actions("jac_misc.cluster")
        assert ret == True

    @jac_testcase("cluster.jac", "test_umap")
    def test_umap(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(len(ret["report"][0]), 15)
        self.assertEqual(len(ret["report"][0][0]), 2)

    @jac_testcase("cluster.jac", "test_get_cluster_hbdscan")
    def test_get_clusters_hbdscan(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(len(ret["report"][0]), 15)

    @jac_testcase("cluster.jac", "test_get_cluster_kmeans")
    def test_get_clusters_kmeans(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(len(ret["report"][0]), 15)

    @classmethod
    def tearDownClass(cls):
        super(TextClusterModule, cls).tearDownClass()
        ret = unload_module("jac_misc.cluster.cluster")
        assert ret == True



import json
import umap
import hdbscan

import numpy as np

import sklearn.cluster as cluster

from jaseci.jsorc.live_actions import jaseci_action
from jaseci.jsorc.remote_actions import launch_server
from fastapi import HTTPException


def get_umap_embedds(
    text_embeddings: list, n_neighbors=15, min_dist=0.1, n_components=2, random_state=42
):
    """
    Dimentionality reduction using umap
    Parameters:
    -----------
    text_embeddings: json object of embeddings.
    n_neighbors: integer
    min_dist: float
    n_components: integer
    random_state: integer
    Return:
    -----------
    clusterable_embedds: ndarray, dimentionality reduced (by umap algorithm) multidimentional array
    """
    emb_str = str(text_embeddings)
    emb_array = np.array(json.loads(emb_str))

    clusterable_embedds = umap.UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        n_components=n_components,
        random_state=random_state,
    ).fit_transform(emb_array)

    return clusterable_embedds.tolist()


def hbdscan_clustering(embeddings, min_samples, min_cluster_size):
    """
    Clustering using hbdscan
    Parameters:
    -----------
    embeddings: ndarray, list of embeddings.
    min_samples: integer.
    min_cluster_size: integer.
    Return:
    -----------
    labels: array of integer, containing correspondant class of each input embeddings
    """

    labels = hdbscan.HDBSCAN(
        min_samples=min_samples, min_cluster_size=min_cluster_size
    ).fit_predict(embeddings)

    return labels


def kmeans_clustering(embeddings, n_clusters):
    """
    Clustering using kmeans algorithm.

    Parameters:
    -----------
    embeddings: ndarray, list of embeddings.
    n_clusters: number of clusters.

    Return:
    -----------
    labels: array of integer, containing correspondant class of each input embeddings
    """
    labels = cluster.KMeans(n_clusters=n_clusters).fit_predict(embeddings)
    print(labels)
    return labels


@jaseci_action(act_group=["cluster"], allow_remote=True)
def get_umap(
    text_embeddings: list,
    n_neighbors: int = 15,
    min_dist: float = 0.1,
    n_components: int = 2,
    random_state: int = 42,
):
    try:
        umap_embedds = get_umap_embedds(
            text_embeddings=text_embeddings,
            n_neighbors=n_neighbors,
            min_dist=min_dist,
            n_components=n_components,
            random_state=random_state,
        )
        return umap_embedds
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


VALID_ALGOS = {"hbdscan", "kmeans"}
KMEANS_PARAMS = {"n_clusters"}
HBDSCAN_PARAMS = {"min_samples", "min_cluster_size"}


@jaseci_action(act_group=["cluster"], allow_remote=True)
def get_cluster_labels(
    embeddings: list,
    algorithm: str = "hbdscan",
    min_samples: int = None,
    min_cluster_size: int = None,
    n_clusters: int = None,
):
    if algorithm not in VALID_ALGOS:
        raise ValueError("the algorithm must be one of %r." % VALID_ALGOS)
    elif algorithm == "hbdscan":
        if min_samples is None:
            raise ValueError(
                "If you are using hbdscan algorith min_samples should be defined"
            )
        elif min_cluster_size is None:
            raise ValueError(
                "If you are using hbdscan algorith min_cluster_size should be defined"
            )
        else:
            try:
                cluster_labels = hbdscan_clustering(
                    embeddings=embeddings,
                    min_samples=min_samples,
                    min_cluster_size=min_cluster_size,
                )
                return cluster_labels.tolist()
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))
    elif algorithm == "kmeans":
        if n_clusters is None:
            raise ValueError(
                "If you are using kmeans algorithm n_clusters should be defined"
            )
        else:
            try:
                cluster_labels = kmeans_clustering(
                    embeddings=embeddings,
                    n_clusters=n_clusters,
                )
                return cluster_labels.tolist()
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    print("The text cluster module is up and running.")
    launch_server(port=8000)



from http.client import ImproperConnectionState
from typing import Any, Dict
import warnings
import os
import traceback
from fastapi import HTTPException
import logging

from jaseci.jsorc.live_actions import jaseci_action

from .utils.util import read_yaml, deep_update, save_custom_python
from .inference import InferenceList

warnings.filterwarnings("ignore")


HEAD_NOT_FOUND = "No Active head found. Please create a head first using create_head."
HEAD_LIST_NOT_FOUND = "No Active head list found. Use create_head_list first."


@jaseci_action(act_group=["ph"], allow_remote=True)
def setup():
    global il, list_config
    dirname = os.path.dirname(__file__)
    list_config = read_yaml(os.path.join(dirname, "config.yaml"))
    if os.path.exists("heads/config.yaml") and os.path.exists("heads/custom.py"):
        logging.warning("Found a heads list in the current directory. Loading it ...")
        il = InferenceList(config=read_yaml("heads/config.yaml"))
    else:
        logging.info("No heads list found. Run create_head_list to create one.")
        il = None


### Start of PersonalizedHead Actions ###


@jaseci_action(act_group=["ph"], allow_remote=True)
def create_head_list(config: Dict = None, python: str = None) -> None:
    """
    Create a holder for all the heads
    """
    try:
        global il, list_config
        if config:
            deep_update(list_config, config)
        if python:
            os.makedirs("heads", exist_ok=True)
            save_custom_python(python)
        il = InferenceList(config=list_config)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ph"], allow_remote=True)
def create_head(uuid: str = None, config: Dict = None) -> None:
    """
    Create a personalized head. This will create a new inference engine.
    @param config: new config to be used for the head
    """
    try:
        global il, list_config
        if config:
            head_config = list_config.copy()
            deep_update(head_config, config)
            _uuid = il.add(config=head_config, uuid=uuid)
        else:
            _uuid = il.add(uuid=uuid)
        return _uuid
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ph"], allow_remote=True)
def predict(uuid: str, data: Any) -> Any:
    """
    Predict using the current active model.
    @param data: data to be used for prediction
    """
    try:
        global il
        if il:
            try:
                return il.predict(uuid, data)
            except ImproperConnectionState:
                raise Exception(HEAD_NOT_FOUND)
            except Exception as e:
                raise e
        else:
            raise Exception(HEAD_LIST_NOT_FOUND)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ph"], allow_remote=True)
def train_head(uuid: str, config: Dict = None, auto_update: bool = True) -> None:
    """
    Train the current active model.
    @param config: new config to be used for training
    """
    try:
        global il
        if il:
            try:
                il.train(uuid, config, auto_update)
            except ImproperConnectionState:
                raise Exception(HEAD_NOT_FOUND)
            except Exception as e:
                raise e
        else:
            raise Exception(HEAD_LIST_NOT_FOUND)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ph"], allow_remote=True)
def get_config(uuid: str) -> Dict:
    """
    Get the config of the personalized head.
    """
    try:
        global il
        if il:
            try:
                return il.get_config(uuid)
            except ImproperConnectionState:
                raise Exception(HEAD_NOT_FOUND)
            except Exception as e:
                raise e
        else:
            raise Exception(HEAD_LIST_NOT_FOUND)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["ph"], allow_remote=True)
def check_head(uuid: str) -> bool:
    try:
        global il
        if il:
            return il.check(uuid)
        else:
            raise Exception(HEAD_LIST_NOT_FOUND)
    except Exception as e:
        print(traceback.format_exc())
        raise HTTPException(status_code=500, detail=str(e))


### End of PersonalizedHead Actions ###


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



import importlib
import pandas as pd
import torch


from .base import BaseDataLoader
from torch.utils.data.dataloader import default_collate


class SnipsDataLoader(BaseDataLoader):
    def __init__(
        self, train_json, batch_size, shuffle=True, validation_split=0.0, num_workers=1
    ):
        df = pd.read_json(train_json)
        self.dataset = SnipsDataset(df)
        super().__init__(
            self.dataset, batch_size, shuffle, validation_split, num_workers
        )


class SnipsDataset(torch.utils.data.Dataset):
    def __init__(self, df):
        self.df = df
        x = self.df["input_ids"]
        y = self.df["label"]
        self.x = torch.tensor(x)
        self.y = torch.tensor(y)

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx):
        return self.x[idx], self.y[idx]


class CustomDataLoader(BaseDataLoader):
    def __init__(
        self, batch_size, shuffle=True, validation_split=0.0, num_workers=1, **kwargs
    ):
        # import the python file
        spec = importlib.util.spec_from_file_location("module.name", "heads/custom.py")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        # get the model
        self.dataset = getattr(module, "CustomDataset")(**kwargs)
        try:
            collate_fn = getattr(module, "collate_fn")
        except AttributeError:
            collate_fn = default_collate
        super().__init__(
            self.dataset, batch_size, shuffle, validation_split, num_workers, collate_fn
        )



from typing import Any, Dict
import torch.nn as nn
import numpy as np
from abc import abstractmethod
from numpy import inf
import torch
from pathlib import Path
from datetime import datetime
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torch.utils.data.sampler import SubsetRandomSampler
from collections import OrderedDict
import os
import shutil

from .logger import TensorboardWriter, get_logger
from . import model as model_module


class BaseModel(nn.Module):
    """
    Base class for all models
    """

    @abstractmethod
    def forward(self, *inputs):
        """
        Forward pass logic
        :return: Model output
        """
        raise NotImplementedError

    def __str__(self):
        """
        Model prints with number of trainable parameters
        """
        model_parameters = filter(lambda p: p.requires_grad, self.parameters())
        params = sum([np.prod(p.size()) for p in model_parameters])
        return super().__str__() + "\nTrainable parameters: {}".format(params)


class BaseTrainer:
    """
    Base class for all trainers
    """

    def __init__(
        self,
        model,
        criterion,
        metric_ftns,
        optimizer,
        config,
        uuid,
        resume=None,
    ):
        self.config = config
        self.logger = get_logger("train")

        save_dir = Path("heads")
        self.run_id = datetime.now().strftime(r"%m%d_%H%M%S")

        self.checkpoint_dir = save_dir / uuid / "runs" / self.run_id
        self.log_dir = save_dir / uuid / "runs" / self.run_id / "logs"
        self.checkpoint_dir.mkdir(parents=True, exist_ok=True)
        self.log_dir.mkdir(parents=True, exist_ok=True)

        self.logger = get_logger("trainer", config["trainer"]["verbosity"])

        self.model = model
        self.criterion = criterion
        self.metric_ftns = metric_ftns
        self.optimizer = optimizer

        cfg_trainer = config["trainer"]
        self.epochs = cfg_trainer["epochs"]
        self.save_period = cfg_trainer["save_period"]
        self.monitor = cfg_trainer.get("monitor", "off")

        # configuration to monitor model performance and save best
        if self.monitor == "off":
            self.mnt_mode = "off"
            self.mnt_best = 0
        else:
            self.mnt_mode, self.mnt_metric = self.monitor.split()
            assert self.mnt_mode in ["min", "max"]

            self.mnt_best = inf if self.mnt_mode == "min" else -inf
            self.early_stop = cfg_trainer.get("early_stop", inf)
            if self.early_stop <= 0:
                self.early_stop = inf

        self.start_epoch = 1

        # setup visualization writer instance
        self.writer = TensorboardWriter(
            self.log_dir, self.logger, cfg_trainer["tensorboard"]
        )

        if resume is not None:
            self._resume_checkpoint(resume)

    @abstractmethod
    def _train_epoch(self, epoch):
        """
        Training logic for an epoch
        :param epoch: Current epoch number
        """
        raise NotImplementedError

    def train(self):
        """
        Full training logic
        """
        not_improved_count = 0
        for epoch in range(self.start_epoch, self.epochs + 1):
            result = self._train_epoch(epoch)

            # save logged informations into log dict
            log = {"epoch": epoch}
            log.update(result)

            # print logged informations to the screen
            for key, value in log.items():
                self.logger.info("    {:15s}: {}".format(str(key), value))

            # evaluate model performance according to configured metric, save best checkpoint as model_best
            best = False
            if self.mnt_mode != "off":
                try:
                    # check whether model performance improved or not, according to specified metric(mnt_metric)
                    improved = (
                        self.mnt_mode == "min" and log[self.mnt_metric] <= self.mnt_best
                    ) or (
                        self.mnt_mode == "max" and log[self.mnt_metric] >= self.mnt_best
                    )
                except KeyError:
                    self.logger.warning(
                        "Warning: Metric '{}' is not found. "
                        "Model performance monitoring is disabled.".format(
                            self.mnt_metric
                        )
                    )
                    self.mnt_mode = "off"
                    improved = False

                if improved:
                    self.mnt_best = log[self.mnt_metric]
                    not_improved_count = 0
                    best = True
                else:
                    not_improved_count += 1

                if not_improved_count > self.early_stop:
                    self.logger.info(
                        "Validation performance didn't improve for {} epochs. "
                        "Training stops.".format(self.early_stop)
                    )
                    break

            if epoch % self.save_period == 0:
                self._save_checkpoint(epoch, save_best=best)

    def _save_checkpoint(self, epoch, save_best=False):
        """
        Saving checkpoints
        :param epoch: current epoch number
        :param log: logging information of the epoch
        :param save_best: if True, rename the saved checkpoint to 'model_best.pth'
        """
        arch = type(self.model).__name__
        state = {
            "arch": arch,
            "epoch": epoch,
            "state_dict": self.model.state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "monitor_best": self.mnt_best,
            "config": self.config,
        }
        filename = str(self.checkpoint_dir / "checkpoint-epoch{}.pth".format(epoch))
        torch.save(state, filename)
        self.logger.info("Saving checkpoint: {} ...".format(filename))
        if save_best:
            best_path = str(self.checkpoint_dir / "model_best.pth")
            torch.save(state, best_path)
            self.logger.info("Saving current best: model_best.pth ...")

    def _resume_checkpoint(self, resume_path):
        """
        Resume from saved checkpoints
        :param resume_path: Checkpoint path to be resumed
        """
        resume_path = str(resume_path)
        self.logger.info("Loading checkpoint: {} ...".format(resume_path))
        checkpoint = torch.load(resume_path, map_location=torch.device("cpu"))
        self.start_epoch = checkpoint.get("epoch", 0) + 1
        self.epochs += checkpoint.get("epoch", 0)
        self.mnt_best = checkpoint.get("monitor_best", 0)

        # load architecture params from checkpoint.
        state_dict = checkpoint.get("state_dict", checkpoint)
        model_keys = list(self.model.state_dict().keys())
        if model_keys[0].startswith("model."):  # when using CustomModel
            new_state_dict = OrderedDict()
            for k, v in state_dict.items():
                name = "model." + k
                new_state_dict[name] = v
            state_dict = new_state_dict
        self.model.load_state_dict(state_dict)

        # load optimizer state from checkpoint only when optimizer type is not changed.
        if checkpoint.get("config", None) is not None:
            if (
                checkpoint["config"]["optimizer"]["type"]
                != self.config["optimizer"]["type"]
            ):
                self.logger.warning(
                    "Warning: Optimizer type given in config file is different from that of checkpoint. "
                    "Optimizer parameters not being resumed."
                )
            else:
                self.optimizer.load_state_dict(checkpoint["optimizer"])

        self.logger.info(
            "Checkpoint loaded. Resume training from epoch {}".format(self.start_epoch)
        )


class BaseDataLoader(DataLoader):
    """
    Base class for all data loaders
    """

    def __init__(
        self,
        dataset,
        batch_size,
        shuffle,
        validation_split,
        num_workers,
        collate_fn=default_collate,
    ):
        self.validation_split = validation_split
        self.shuffle = shuffle

        self.batch_idx = 0
        self.n_samples = len(dataset)

        self.sampler, self.valid_sampler = self._split_sampler(self.validation_split)

        self.init_kwargs = {
            "dataset": dataset,
            "batch_size": batch_size,
            "shuffle": self.shuffle,
            "collate_fn": collate_fn,
            "num_workers": num_workers,
        }
        super().__init__(sampler=self.sampler, **self.init_kwargs)

    def _split_sampler(self, split):
        if split == 0.0:
            return None, None

        idx_full = np.arange(self.n_samples)

        np.random.seed(0)
        np.random.shuffle(idx_full)

        if isinstance(split, int):
            assert split > 0
            assert (
                split < self.n_samples
            ), "validation set size is configured to be larger than entire dataset."
            len_valid = split
        else:
            len_valid = int(self.n_samples * split)

        valid_idx = idx_full[0:len_valid]
        train_idx = np.delete(idx_full, np.arange(0, len_valid))

        train_sampler = SubsetRandomSampler(train_idx)
        valid_sampler = SubsetRandomSampler(valid_idx)

        # turn off shuffle option which is mutually exclusive with sampler
        self.shuffle = False
        self.n_samples = len(train_idx)

        return train_sampler, valid_sampler

    def split_validation(self):
        if self.valid_sampler is None:
            return None
        else:
            return DataLoader(sampler=self.valid_sampler, **self.init_kwargs)


class BaseInference:
    def __init__(self, config: Dict, logger, uuid: str = None) -> None:
        self.config = config
        model_config = self.config["Model"]
        self.infer_config = self.config["Inference"]

        self.model = getattr(model_module, model_config["type"])(
            **model_config.get("args", {})
        )

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # loading pretrained weights
        if self.infer_config["weights"]:
            if not os.path.exists(f"heads/{uuid}/current.pth"):
                shutil.copyfile(
                    self.infer_config["weights"], f"heads/{uuid}/current.pth"
                )
                logger.info(
                    "Loading default checkpoint: {} ...".format(
                        self.infer_config["weights"]
                    )
                )

            checkpoint = torch.load(
                f"heads/{uuid}/current.pth", map_location=torch.device("cpu")
            )
            state_dict = checkpoint.get("state_dict", checkpoint)
            model_keys = list(self.model.state_dict().keys())
            if model_keys[0].startswith("model."):
                new_state_dict = OrderedDict()
                for k, v in state_dict.items():
                    name = "model." + k
                    new_state_dict[name] = v
                state_dict = new_state_dict
            self.model.load_state_dict(state_dict)
            logger.info("Checkpoint loaded.")

        # Setting the device
        self.model = self.model.to(self.device)
        self.model.eval()

        # Get Intermediate Activations
        self.out_activation_layer = self.infer_config.get("out_activation_layer", None)
        if self.out_activation_layer:
            self.activation = {}
            getattr(self.model, self.out_activation_layer).register_forward_hook(
                self.get_activation(self.out_activation_layer)
            )

    @torch.no_grad()
    def predict(self, data: Any) -> Any:
        data = self.preprocess(data)
        data = data.to(self.device)
        output = self.model(data)
        if self.out_activation_layer:
            output = self.activation[self.out_activation_layer]
        output = self.postprocess(output)
        return output

    def preprocess(self, data: Any) -> Any:
        raise NotImplementedError

    def postprocess(self, data: Any) -> Any:
        raise NotImplementedError

    def load_weights(self, weights: str) -> None:
        checkpoint = torch.load(weights, map_location=torch.device("cpu"))
        state_dict = checkpoint.get("state_dict", checkpoint)
        self.model.load_state_dict(state_dict)

    def get_activation(self, name):
        def hook(model, input, output):
            self.activation[name] = output.detach()

        return hook



import torch
from transformers import AutoTokenizer, AutoModel
import importlib


class SnipsPostProcessor:
    def __init__(self, to_list: bool = False):
        self.to_list = to_list

    def process(self, output):
        output = torch.argmax(output, dim=1)
        if self.to_list:
            output = output.tolist()[0]
        return output


class SnipsPreProcessor:
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")
        self.model = AutoModel.from_pretrained("bert-base-uncased")

    def process(self, input):
        input = self.tokenizer(input, return_tensors="pt")
        emb = self.model(**input)
        emb = emb[0][:, 0, :]
        return emb


class V2VPostProcessor:
    def __init__(self):
        pass

    def process(self, output: torch.Tensor) -> list:
        return output.tolist()


class V2VPreProcessor:
    def __init__(self):
        pass

    def process(self, input: list) -> torch.Tensor:
        input = torch.tensor(input)
        return input


class CustomProcessor:
    def __init__(
        self, module_name: str, python_file: str = "heads/custom.py", **kwargs
    ):
        spec = importlib.util.spec_from_file_location("module.name", python_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        self.processor = getattr(module, module_name)(**kwargs)

    def process(self, input):
        return self.processor.process(input)



from http.client import ImproperConnectionState
import importlib
from typing import Any, Dict
import torch
import uuid as uuid_gen
import os
import shutil
from collections import OrderedDict
import json

from .utils import model as model_module
from .utils import process as process_module
from .utils.base import BaseInference
from .utils.logger import get_logger
import logging
from .utils.util import deep_update, write_yaml, read_yaml

from .train import train

import warnings

warnings.filterwarnings("ignore")


class InferenceEngine:
    """
    Inference Engine for each user.
    @param: config: Dict
    """

    def __init__(self, config: Dict, uuid=None):
        self.ph_config = config
        self.id = uuid if uuid else str(uuid_gen.uuid4())

        os.makedirs(f"heads/{self.id}", exist_ok=True)
        write_yaml(self.ph_config, f"heads/{self.id}/config.yaml")

        self.logger = get_logger(name=self.id)
        self.logger.addHandler(logging.FileHandler(f"heads/{self.id}/log.txt"))
        self.logger.setLevel(logging.DEBUG)

        # Creating the Inference Object
        if config["Inference"].get("type", None) == "CustomInference":
            spec = importlib.util.spec_from_file_location(
                "module.name", "heads/custom.py"
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            self.pipeline = getattr(module, "CustomInference")(
                self.ph_config, self.logger, self.id
            )
        else:
            self.pipeline = BaseInference(self.ph_config, self.logger, self.id)
            preprocessor_args = self.ph_config["Inference"]["preprocess"].get(
                "args", {}
            )
            if self.ph_config["Inference"]["preprocess"]["type"].startswith("Custom"):
                preprocessor_args["module_name"] = "CustomPreProcessor"
            preprocessor = getattr(
                process_module, self.ph_config["Inference"]["preprocess"]["type"]
            )(**preprocessor_args)
            self.pipeline.preprocess = preprocessor.process

            postprocessor_args = self.ph_config["Inference"]["postprocess"].get(
                "args", {}
            )
            if self.ph_config["Inference"]["postprocess"]["type"].startswith("Custom"):
                postprocessor_args["module_name"] = "CustomPostProcessor"
            postprocessor = getattr(
                process_module, self.ph_config["Inference"]["postprocess"]["type"]
            )(**postprocessor_args)
            self.pipeline.postprocess = postprocessor.process

    @torch.no_grad()
    def predict(self, data: Any) -> Any:
        self.logger.info("Predict is called")
        return self.pipeline.predict(data)

    def load_weights(self, weights: str) -> None:
        self.logger.info("Loading new weights: {} ...".format(weights))
        self.pipeline.load_weights(weights)

    def __del__(self):
        del self.pipeline
        torch.cuda.empty_cache()


class InferenceList:
    """
    Parent of Inference Engines. use to manage multiple Inference Engines.
    """

    def __init__(self, config: Dict = None) -> None:  # type: ignore
        self.config = config
        with open("heads/config.yaml", "w") as f:
            json.dump(config, f)
        os.makedirs("heads", exist_ok=True)
        self.ie_list = {}
        self.ph_list = []
        if os.path.exists("heads/ph_list.json"):
            with open("heads/ph_list.json", "r") as f:
                self.ph_list = json.load(f)

    def add(self, config: Dict = None, uuid: str = None) -> str:  # type: ignore
        if self.check(uuid):
            raise ImproperConnectionState(
                f"{self.check(uuid)} Inference Engine already exists. Please use another uuid."
            )
        if config:
            ie = InferenceEngine(config, uuid)
        else:
            ie = InferenceEngine(self.config, uuid)
        self.ie_list[ie.id] = ie
        if not os.path.exists("heads/ph_list.json"):
            with open("heads/ph_list.json", "w") as f:
                json.dump([], f)
        with open("heads/ph_list.json", "r") as f:
            self.ph_list = json.load(f)
            self.ph_list.append(ie.id)
        with open("heads/ph_list.json", "w") as f:
            json.dump(self.ph_list, f)
        return ie.id

    def predict(self, uuid: str, data: Any) -> Any:
        if self.check(uuid):
            return self.ie_list[uuid].predict(data)
        else:
            raise ImproperConnectionState("Inference Engine not found.")

    def train(self, uuid: str, config: Dict = None, auto_update=True) -> None:  # type: ignore
        if self.check(uuid):
            ph_config = self.ie_list[uuid].ph_config
            if config:
                # If the training data is passed in, save it to a file first
                dataset = config["Trainer"]["dataloader"]["args"].get("data_dict", None)
                if dataset:
                    data_loc = config["Trainer"]["dataloader"]["args"].get("data_dir")
                    with open(data_loc, "w") as fout:
                        json.dump(dataset, fout)
                    del config["Trainer"]["dataloader"]["args"]["data_dict"]

                deep_update(ph_config, config)
                write_yaml(ph_config, f"heads/{uuid}/config.yaml")
            resume = (
                f"heads/{uuid}/current.pth"
                if os.path.exists(f"heads/{uuid}/current.pth")
                else None
            )

            run_id = train(
                {
                    "uuid": uuid,
                    "config": f"heads/{uuid}/config.yaml",
                    "resume": resume,
                    "device": None,
                }
            )
            if auto_update:
                self.update_head(uuid, run_id)
        else:
            raise ImproperConnectionState("Inference Engine not found.")

    def update_head(self, uuid: str, run_id: str) -> None:
        if os.path.exists(f"heads/{uuid}/runs/{run_id}/model_best.pth"):
            shutil.copyfile(
                f"heads/{uuid}/runs/{run_id}/model_best.pth",
                f"heads/{uuid}/current.pth",
            )
            self.ie_list[uuid].load_weights(f"heads/{uuid}/current.pth")
        else:
            print(f"Model for {uuid} has not improved. No update.")

    def get_config(self, uuid: str) -> Dict:
        if self.check(uuid):
            return self.ie_list[uuid].ph_config
        else:
            raise ImproperConnectionState("Inference Engine not found.")

    def delete_head(self, uuid: str) -> None:
        if self.check(uuid):
            del self.ie_list[uuid]
            self.ph_list.remove(uuid)
            with open("heads/ph_list.json", "w") as f:
                json.dump(self.ph_list, f)
            shutil.rmtree(f"heads/{uuid}")
        else:
            raise ImproperConnectionState("Inference Engine not found.")

    def load_head(self, uuid: str) -> None:
        config = read_yaml(f"heads/{uuid}/config.yaml")
        self.ie_list[uuid] = InferenceEngine(config, uuid)
        if os.path.exists(f"heads/{uuid}/current.pth"):
            self.ie_list[uuid].load_weights(f"heads/{uuid}/current.pth")

    def check(self, uuid: str) -> bool:
        if uuid in self.ph_list and uuid in self.ie_list:
            return True
        if uuid in self.ph_list:
            self.load_head(uuid)
            return True
        return False



"""
OpenAI Integration for Jaseci
Features:
    - GPT-3 (Completion and Chat)
    - Codex (Code Completion)
    - DALL-E (Image Generation)
    - Whispers (Audio Transcription)

Requires OpenAI API Key
"""

from jaseci.jsorc.live_actions import jaseci_action
from jaseci.utils.utils import logger
import os
import openai
from typing import Union


@jaseci_action(act_group=["openai"], allow_remote=True)
def setup(api_key=os.environ.get("OPENAI_API_KEY", None)):
    """
    Sets up OpenAI API Key

    :param api_key: str, required, default=None
                   OpenAI API key
    :return: bool
             Returns True if api_key is set, False otherwise
    """
    if not api_key:
        logger.error(
            "No OpenAI API Key Provided. Please set OPENAI_API_KEY environment variable or pass in api_key parameter though actions call openai.setup"
        )
        return False
    else:
        openai.api_key = api_key
        return True


@jaseci_action(act_group=["openai"], allow_remote=True)
def completion(
    prompt: Union[str, list] = "<|endoftext|>",
    model: str = "text-davinci-003",
    suffix: str = None,
    max_tokens: int = 16,
    temperature: float = 1,
    top_p: float = 1,
    n: int = 1,
    logprobs: int = None,
    echo: bool = False,
    stop: Union[str, list] = None,
    presence_penalty: float = 0,
    frequency_penalty: float = 0,
    best_of: int = 1,
):
    """
    Generate text completions based on a prompt using OpenAI's GPT-3 models.

    Parameters:
    ----------
    prompt : str or list of str, optional (default="")
        The prompt(s) to generate text completions for.
    model : str, optional (default="text-davinci-003")
        The name of the GPT-3 model to use for generating completions.
    suffix : str, optional (default=None)
        A suffix to append to the completion(s).
    max_tokens : int, optional (default=16)
        The maximum number of tokens to generate for each completion.
    temperature : float, optional (default=1.0)
        Controls the randomness of the generated completions. Higher values will result in more varied completions.
    top_p : float, optional (default=1.0)
        Controls the diversity of the generated completions. Lower values will result in more conservative completions.
    n : int, optional (default=1)
        The number of completions to generate.
    logprobs : int or None, optional (default=None)
        Controls whether to include log probabilities with the generated completions. If set to an integer value, the top n log probabilities for each token will be returned.
    echo : bool, optional (default=False)
        Controls whether the prompt should be included in the generated completions.
    stop : str, list of str, or None, optional (default=None)
        The sequence at which the model should stop generating text. If a list is provided, the model will stop at any of the specified sequences.
    presence_penalty : float, optional (default=0.0)
        Controls the model's tendency to generate new words or phrases. Higher values will result in more novel completions.
    frequency_penalty : float, optional (default=0.0)
        Controls the model's tendency to repeat words or phrases. Higher values will result in less repetitive completions.
    best_of : int, optional (default=1)
        Controls how many completions to return, and selects the best one(s) based on their log probabilities.

    Returns:
    -------
    completions : list of str
        A list of completions generated by the GPT-3 model based on the provided prompt(s).
    """

    response = openai.Completion.create(
        model=model,
        prompt=prompt,
        suffix=suffix,
        max_tokens=max_tokens,
        temperature=temperature,
        top_p=top_p,
        n=n,
        logprobs=logprobs,
        echo=echo,
        stop=stop,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
        best_of=best_of,
    )
    response = [x.text for x in response.choices]
    return response


@jaseci_action(act_group=["openai"], allow_remote=True)
def chat(
    messages: list,
    model: str = "gpt-3.5-turbo",
    temperature: float = 1,
    top_p: float = 1,
    n: int = 1,
    stop: Union[str, list] = None,
    presence_penalty: float = 0,
    frequency_penalty: float = 0,
):
    """
    Generate responses to a list of messages using OpenAI's GPT-3.5 model.

    Parameters:
    ----------
    messages : list of str
        A list of messages to prompt the GPT-3.5 model with.
    model : str, optional (default='gpt-3.5-turbo')
        The name of the GPT-3.5 model to use for generating responses.
    temperature : float, optional (default=1.0)
        Controls the randomness of the generated responses. Higher values will result in more varied responses.
    top_p : float, optional (default=1.0)
        Controls the diversity of the generated responses. Lower values will result in more conservative responses.
    n : int, optional (default=1)
        The number of responses to generate for each message.
    stop : str, list of str, or None, optional (default=None)
        The sequence at which the model should stop generating text. If a list is provided, the model will stop at any of the specified sequences.
    presence_penalty : float, optional (default=0.0)
        Controls the model's tendency to generate new words or phrases. Higher values will result in more novel responses.
    frequency_penalty : float, optional (default=0.0)
        Controls the model's tendency to repeat words or phrases. Higher values will result in less repetitive responses.

    Returns:
    -------
    responses : list of str
        A list of responses generated by the GPT-3.5 model based on the provided messages.
    """

    response = openai.ChatCompletion.create(
        model=model,
        prompt=messages,
        temperature=temperature,
        top_p=top_p,
        n=n,
        stop=stop,
        presence_penalty=presence_penalty,
        frequency_penalty=frequency_penalty,
    )
    response = [x.message for x in response.choices]
    return response


@jaseci_action(act_group=["openai"], allow_remote=True)
def get_embeddings(input: Union[str, list], model: str = "text-embedding-ada-002"):
    """
    Returns the embedding of the input text(s) using the specified OpenAI embedding model.

    Parameters:
    -----------
    input: Union[str, list]
        The input text or a list of input texts to be embedded.
    model: str, optional (default="text-embedding-ada-002")
        The OpenAI embedding model to use.

    Returns:
    --------
    List[List[float]]:
        A list of embeddings, where each embedding is a list of floating-point numbers.
    """
    response = openai.Embedding.create(model=model, inputs=input)
    response = [x.embedding for x in response.data]
    return response


@jaseci_action(act_group=["openai"], allow_remote=True)
def transcribe(
    audio_file: str = None,
    audio_url: str = None,
    audio_array: list = None,
    model: str = "whisper-1",
    prompt: str = None,
    temperature: float = 0,
    language: str = None,
    translate=False,
):
    """
    Transcribes an audio file, audio URL, or audio array using the specified OpenAI Whisper model.

    Parameters:
    -----------
    audio_file: str, optional
        The path to the audio file to transcribe.
    audio_url: str, optional
        The URL of the audio file to transcribe.
    audio_array: list, optional
        A list containing the audio waveform as a sequence of floats between -1 and 1.
    model: str, optional (default="whisper-1")
        The OpenAI Whisper model to use for transcription.
    prompt: str, optional
        A prompt to use in addition to the audio input when transcribing with the OpenAI API.
    temperature: float, optional (default=0)
        The temperature to use when generating text for the transcription.
    language: str, optional
        The language of the audio being transcribed, specified as a BCP-47 language code.
    translate: bool, optional (default=False)
        If True, the transcribed text will be translated to English.

    Returns:
    --------
    str:
        The transcribed text.
    """
    if audio_array:
        import soundfile as sf
        import tempfile
        import numpy as np

        audio_array = np.array(audio_array)

        with tempfile.NamedTemporaryFile(suffix=".wav") as fp:
            sf.write(fp.name, audio_array, 16000)
            audio_file = fp.name
    if audio_url:
        import urllib.request
        import tempfile

        with tempfile.NamedTemporaryFile(suffix=".wav") as fp:
            urllib.request.urlretrieve(audio_url, fp.name)
            audio_file = fp.name

    audio_file = open(audio_file, "rb")
    if not translate:
        transcription = openai.Audio.transcribe(
            file=audio_file,
            model=model,
            prompt=prompt,
            temperature=temperature,
            language=language,
        )
    else:
        transcription = openai.Audio.translate(
            file=audio_file,
            model=model,
            prompt=prompt,
            temperature=temperature,
            language=language,
            translate=translate,
        )
    return transcription.text


@jaseci_action(act_group=["openai"], allow_remote=True)
def generate_image(
    prompt: str, n: int = 1, size: str = "512x512", response_format: str = "url"
):
    """
    Generate images using the DALL-E 2 API.

    Parameters:
    -----------
    prompt : str
        The textual prompt for generating the image(s).
    n : int, optional
        The number of images to generate (default is 1).
    size : str, optional
        The size of the image in pixels (default is "512x512").
    response_format : str, optional
        The format of the response, either "url" or "json" (default is "url").

    Returns:
    --------
    A list of generated images, either as URLs or Base64-encoded JSON strings,
    depending on the value of `response_format`.
    """
    response = openai.Image.create(
        prompt=prompt, n=n, size=size, response_format=response_format
    )
    response = [
        x.url if response_format == "url" else x.b64_json for x in response.data
    ]
    return response


@jaseci_action(act_group=["openai"], allow_remote=True)
def edit_image(
    prompt: str,
    image_file: str = None,
    mask_file: str = None,
    image_b64: str = None,
    mask_b64: str = None,
    n: int = 1,
    size: str = "512x512",
    response_format: str = "url",
):
    """
    DALL-E 2 Image Edit API.

    Args:
        prompt (str): Prompt describing the editing task.
        image_file (str, optional): Path to the input image file. Defaults to None.
        mask_file (str, optional): Path to the mask image file. Defaults to None.
        image_b64 (str, optional): Base64 encoded input image. Defaults to None.
        mask_b64 (str, optional): Base64 encoded mask image. Defaults to None.
        n (int, optional): Number of images to generate. Defaults to 1.
        size (str, optional): Size of the generated image. Defaults to "512x512".
        response_format (str, optional): Format of the response. "url" or "base64". Defaults to "url".

    Returns:
        List[str]: List of URLs or base64 encoded images.
    """
    if image_b64:
        import base64
        import tempfile

        image = base64.b64decode(image_b64)
        with tempfile.NamedTemporaryFile(suffix=".png") as fp:
            fp.write(image)
            image_file = fp.name
    if mask_b64:
        import base64
        import tempfile

        mask = base64.b64decode(mask_b64)
        with tempfile.NamedTemporaryFile(suffix=".png") as fp:
            fp.write(mask)
            mask_file = fp.name

    image = open(image_file, "rb")
    mask = open(mask_file, "rb")

    response = openai.Image.create_edit(
        prompt=prompt,
        n=n,
        size=size,
        response_format=response_format,
        image=image,
        mask=mask,
    )
    response = [
        x.url if response_format == "url" else x.b64_json for x in response.data
    ]
    return response


@jaseci_action(act_group=["openai"], allow_remote=True)
def variations_image(
    image_file: str = None,
    image_b64: str = None,
    n: int = 1,
    size: str = "512x512",
    response_format: str = "url",
):
    """
    Generates n variations of an image using the DALL-E 2 Image Variation API.

    :param image_file: (Optional) Path to the image file to use. If not provided, `image_b64` must be provided instead.
    :type image_file: str
    :param image_b64: (Optional) Base64-encoded image data. If not provided, `image_file` must be provided instead.
    :type image_b64: str
    :param n: (Optional) Number of variations to generate. Default is 1.
    :type n: int
    :param size: (Optional) Size of the output images in the format "<width>x<height>". Default is "512x512".
    :type size: str
    :param response_format: (Optional) Format of the response data. Can be "url" or "b64_json". Default is "url".
    :type response_format: str

    :return: A list of URLs or base64-encoded JSON strings representing the generated images.
    :rtype: List[str]
    """
    if image_b64:
        import base64
        import tempfile

        image = base64.b64decode(image_b64)
        with tempfile.NamedTemporaryFile(suffix=".png") as fp:
            fp.write(image)
            image_file = fp.name

    image = open(image_file, "rb")

    response = openai.Image.create_variation(
        n=n, size=size, response_format=response_format, image=image
    )
    response = [
        x.url if response_format == "url" else x.b64_json for x in response.data
    ]
    return response



import importlib.util
import torch.nn as nn
import torch.nn.functional as F
import torch

from .base import BaseModel

import warnings

warnings.filterwarnings("ignore")


class PHClassifier(BaseModel):
    def __init__(
        self, embedding_length, ph_nhead, ph_ff_dim, batch_first, ph_nlayers, n_classes
    ):
        super().__init__()
        encoder_layer = nn.TransformerEncoderLayer(
            d_model=embedding_length,
            nhead=ph_nhead,
            dim_feedforward=ph_ff_dim,
            batch_first=batch_first,
        )
        self.encoder = nn.TransformerEncoder(
            encoder_layer=encoder_layer, num_layers=ph_nlayers
        )
        self.decoder = nn.Linear(embedding_length, n_classes)
        nn.init.xavier_uniform_(self.decoder.weight)

    def forward(self, emb):
        x = self.encoder(emb)
        x = self.decoder(x)
        x = F.log_softmax(x, dim=-1)
        return x.squeeze(1)


class PHVector2Vector(BaseModel):
    def __init__(
        self,
        input_dim,
        ph_nhead,
        ph_ff_dim,
        batch_first,
        ph_nlayers,
        output_dim,
    ):
        super().__init__()
        encoder_layer = nn.TransformerEncoderLayer(
            d_model=input_dim,
            nhead=ph_nhead,
            dim_feedforward=ph_ff_dim,
            batch_first=batch_first,
        )
        self.encoder = nn.TransformerEncoder(
            encoder_layer=encoder_layer, num_layers=ph_nlayers
        )
        self.decoder = nn.Linear(input_dim, output_dim)
        nn.init.xavier_uniform_(self.decoder.weight)

    def forward(self, emb):
        x = self.encoder(emb)
        x = self.decoder(x)
        return x


class PHVectorSimilarity(BaseModel):
    def __init__(
        self, input_dim, ph_nhead, ph_ff_dim, batch_first, ph_nlayers, output_dim
    ):
        super().__init__()
        self.input_dim = input_dim
        self.ph_v2v = PHVector2Vector(
            input_dim, ph_nhead, ph_ff_dim, batch_first, ph_nlayers, output_dim
        )

    def forward(self, emb):
        if emb.shape[1] == self.input_dim * 2:
            emb1 = emb[:, : self.input_dim]
            emb2 = emb[:, self.input_dim :]
            assert emb1.shape == emb2.shape
            x1 = self.ph_v2v(emb1)
            x2 = self.ph_v2v(emb2)
            x = torch.cat((x1, x2), dim=1)
        else:
            assert emb.shape[1] == self.input_dim
            x = self.ph_v2v(emb)
        return x


class CustomModel(BaseModel):
    def __init__(self, python_file: str = "heads/custom.py", **kwargs):
        super().__init__()
        # import the python file
        spec = importlib.util.spec_from_file_location("module.name", python_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        # get the model
        self.model = getattr(module, "CustomModel")(**kwargs)

    def forward(self, x):
        x = self.model(x)
        return x



import torch.nn.functional as F
import torch
import importlib.util


def nll_loss(output, target):
    loss = F.nll_loss(output, target)
    return loss


def cross_entropy_loss(output, target):
    loss = F.cross_entropy(output, target)
    return loss


def mse_loss(output, target):
    loss = F.mse_loss(output, target, reduction="mean")
    return loss


class VectorSimilarityLoss(torch.nn.Module):
    def __init__(self, output_dim=128):
        super(VectorSimilarityLoss, self).__init__()
        self.output_dim = output_dim

    def forward(self, output, target):
        output1, output2 = output[:, : self.output_dim], output[:, self.output_dim :]
        similarity = torch.nn.functional.cosine_similarity(output1, output2, dim=1)
        target = target.float()
        loss = torch.nn.functional.cross_entropy(similarity, target)
        return loss


class CustomLoss(torch.nn.Module):
    def __init__(self, python_file: str = "heads/custom.py", **kwargs):
        super().__init__()
        # import the python file
        spec = importlib.util.spec_from_file_location("module.name", python_file)
        foo = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(foo)  # type: ignore
        self.loss = foo.CustomLoss(**kwargs)

    def forward(self, output, target):
        loss = self.loss(output, target)
        return loss



import requests
import traceback

try:
    with open("custom.py", "r") as fp:
        model_data = fp.read()
    req_json = {
        "config": {
            "Model": {
                "args": {
                    "model_args": {
                        "context_bert_model": "distilbert-base-uncased",
                        "candidate_bert_model": "distilbert-base-uncased",
                        "hidden_size": 128,
                        "max_sequence_length": 128,
                        "max_entity_length": 30,
                        "loss_beta": 0.6,
                        "start_coef": 0.3,
                        "end_coef": 0.1,
                        "span_coef": 0.6,
                        "unk_entity_type_id": -1,
                        "unk_category": "<UNK>",
                        "descriptions": ["LOC"],
                    }
                },
                "type": "CustomModel",
            },
            "Inference": {"type": "CustomInference"},
        },
        "python": model_data,
    }
    response = requests.post("http://localhost:8000/create_head_list/", json=req_json)
    print(response.status_code)
    if response.status_code == 200:
        response = requests.post(
            "http://localhost:8000/create_head/", json={"uuid": "sample", "config": {}}
        )
        print(response.status_code)
        print(response.text)
        if response.status_code == 200:
            train_json = {
                "uuid": "sample",
                "config": {
                    "Trainer": {
                        "name": "CustomTrainer",
                        "trainer": {"epochs": 25, "tensorboard": False},
                        "dataloader": {
                            "args": {
                                "train_args": {
                                    "context_bert_model": "distilbert-base-uncased",
                                    "hidden_size": 128,
                                    "max_sequence_length": 128,
                                    "max_entity_length": 30,
                                    "unk_entity_type_id": -1,
                                    "unk_category": "<UNK>",
                                    "descriptions": ["LOC"],
                                    "data_file": "ph/ph_train_data.json",
                                }
                            },
                            "type": "CustomDataLoader",
                        },
                        "loss": "custom_loss",
                        "loss_args": {"n_classes": 1},
                        "metrics": [],
                    }
                },
            }
            response = requests.post(
                "http://localhost:8000/train_head/", json=train_json
            )
            print(response.status_code)
            print(response.text)
            if response.status_code == 200:
                pred_json = {
                    "uuid": "sample",
                    "data": {
                        "inf_args": {
                            "unk_entity_type_id": -1,
                            "unk_category": "<UNK>",
                            "max_sequence_length": 128,
                            "descriptions": ["PER", "LOC"],
                        },
                        "dataset": [
                            "China is surpassed in area by only Russia",
                            "China also faces Korea and Japan across the Yellow Sea",
                        ],
                    },
                }
                response = requests.post(
                    "http://localhost:8000/predict/", json=pred_json
                )
                print(response.status_code)
                print(response.text)
except Exception as e:
    print(f"Exceptions : {e}")
    print(traceback.print_exc())



import argparse
import torch
import os
from pathlib import Path

from .utils import loss as loss_module
from .utils import metric as metric_module
from .utils import model as model_module
from .utils import dataloader as dataloader_module

from .utils.trainer import Trainer
from .utils import prepare_device, read_yaml
from .utils.logger import setup_logging


def train(args):
    if args["device"] is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.device

    # Loading Configuration
    if args["resume"] is not None:
        resume = Path(args["resume"])
    else:
        resume = None

    msg_no_cfg = (
        "Configuration file need to be specified. Add '-c config.yaml', for example."
    )
    assert args["config"] is not None, msg_no_cfg
    cfg_fname = Path(args["config"])

    config = read_yaml(cfg_fname)

    # Setup Logger
    log_config = config["Logger"]
    setup_logging(log_config)

    # Build model Architecture
    model_config = config["Model"]
    model = getattr(model_module, model_config["type"])(**model_config.get("args", {}))

    trainer_config = config["Trainer"]

    # Setup Dataloader
    dataloader = getattr(dataloader_module, trainer_config["dataloader"]["type"])(
        **trainer_config["dataloader"].get("args", {})
    )
    valid_dataloader = dataloader.split_validation()

    # for multi-GPU training
    device, device_ids = prepare_device(trainer_config["n_gpu"])
    model = model.to(device)
    if len(device_ids) > 1:
        model = torch.nn.DataParallel(model, device_ids=device_ids)

    # get function handles of loss and metrics
    if trainer_config["loss"] == "custom_loss":
        criterion = getattr(loss_module, "CustomLoss")(
            **trainer_config.get("loss_args", {})
        )
    elif trainer_config["loss"] == "vector_similarity_loss":
        criterion = getattr(loss_module, "VectorSimilarityLoss")(
            **trainer_config.get("loss_args", {})
        )
    else:
        criterion = getattr(loss_module, trainer_config["loss"])

    metrics = [getattr(metric_module, metric) for metric in trainer_config["metrics"]]

    # build optimizer, learning rate scheduler. delete every lines containing lr_scheduler for disabling scheduler
    trainable_params = filter(lambda p: p.requires_grad, model.parameters())
    optimizer = getattr(torch.optim, trainer_config["optimizer"]["type"])(
        trainable_params, **trainer_config["optimizer"].get("args", {})
    )
    lr_scheduler = getattr(
        torch.optim.lr_scheduler, trainer_config["lr_scheduler"]["type"]
    )(optimizer, **trainer_config["lr_scheduler"].get("args", {}))

    trainer = Trainer(
        model,
        criterion,
        metrics,
        optimizer,
        config=trainer_config,
        device=device,
        data_loader=dataloader,
        valid_data_loader=valid_dataloader,
        lr_scheduler=lr_scheduler,
        resume=resume,
        uuid=args["uuid"],
    )

    trainer.train()
    return trainer.run_id


if __name__ == "__main__":
    args = argparse.ArgumentParser(description="PyTorch Template")
    args.add_argument(
        "-c",
        "--config",
        default=None,
        type=str,
        help="config file path (default: None)",
    )
    args.add_argument(
        "-r",
        "--resume",
        default=None,
        type=str,
        help="path to latest checkpoint (default: None)",
    )
    args.add_argument(
        "-d",
        "--device",
        default=None,
        type=str,
        help="indices of GPUs to enable (default: all)",
    )
    args.add_argument(
        "-u",
        "--uuid",
        default=None,
        type=str,
        help="uuid of the experiment (default: None)",
    )
    args = args.parse_args()
    # convert args to dict
    args = vars(args)
    train(args)



import requests
import os
import re
import uuid

from jaseci.jsorc.live_actions import jaseci_action
from PyPDF2 import PdfFileReader
from fastapi import HTTPException


def download_pdf(url: str, filename: str):
    r"""method for downloading PDF from URL
    :param url: URL for the downloading a pdf file.
    :param filename: filename for saving downloaded pdf locally.
    """
    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        with open(filename, "wb") as f:
            for chunk in req.iter_content(chunk_size=2048):
                f.write(chunk)


def valid_url(url: str):
    r"""validating whether given url is a valid pdf url or not
    :param url: URL for the downloading a pdf file.
    """
    r = requests.get(url)
    content_type = r.headers.get("content-type")
    return True if "application/pdf" in content_type else False


def valid_file_path(path):
    r"""validating whether given path is valid or not
    :param path: Path of the pdf file"""
    if os.path.exists(path):
        return True
    return False


def remove_pdf(filename: str):
    r"""REMOVE locally downloaded PDF
    :param filename: filename for the deleting pdf file.
    """
    if os.path.exists(filename):
        os.remove(filename)
        return True
    return False


def process_pdf(filename, metadata, data):
    with open(filename, "rb") as pdf_file:
        pdf_reader = PdfFileReader(pdf_file)
        if metadata:
            data.update({"metadata": {}})
            md = dict(pdf_reader.documentInfo)
            for k, v in md.items():
                data["metadata"][re.sub("[^a-zA-Z0-9]+", "", k)] = v
        data["pages"] = len(pdf_reader.pages)
        data["content"] = [page.extractText() for page in pdf_reader.pages]
    return data


@jaseci_action(act_group=["pdf_ext"], allow_remote=True)
def extract_pdf(url: str = None, path: str = None, metadata: bool = False):
    r"""REMOVE locally downloaded PDF
    :param filename: filename for the deleting pdf file.
    :param metadata: boolean if you want to display available metadata of PDF.
    """

    filename = str(uuid.uuid4().hex) + ".pdf"
    data = {"pages": 0, "content": None}
    if url != None:
        if valid_url(url):
            try:
                download_pdf(url, filename)
                data = process_pdf(filename, metadata, data)
                remove_pdf(filename)
                return data
            except Exception:
                remove_pdf(filename)

                raise HTTPException(
                    status_code=500, detail="Unable to extract data from pdf"
                )
        else:
            raise HTTPException(status_code=415, detail=str("Invalid file format"))

    elif path != None:
        if valid_file_path(path):
            data = process_pdf(path, metadata, data)
            return data
        raise HTTPException(status_code=400, detail=str("Invalid path"))
    else:
        raise HTTPException(status_code=400, detail=str("Missing params url/path"))


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



from .action_configs.pdf_ext_action_config import PDF_EXT_ACTION_CONFIG
from .action_configs.cluster_action_config import CLUSTER_ACTION_CONFIG
from .action_configs.example_module_action_config import EXAMPLE_MODULE_ACTION_CONFIG

ACTION_CONFIGS = {
    "pdf_ext": PDF_EXT_ACTION_CONFIG,
    "cluster": CLUSTER_ACTION_CONFIG,
    "example_module": EXAMPLE_MODULE_ACTION_CONFIG,
}
